@@ -15,6 +15,8 @@
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
 import datetime
+import functools
+import itertools
 
 
 class Bug(object):
@@ -31,6 +33,19 @@ class Bug(object):
     def data(self, value):
         self._data = value
 
+    @property
+    def history(self):
+        if self._history is None:
+            if not self.bugno:
+                raise Exception("bugno not provided.")
+            self._history = \
+                self.rpc('history', ids=[self.bugno])['bugs'][0]['history']
+        return self._history
+
+    @history.setter
+    def history(self, value):
+        self._history = value
+
     @property
     def comments(self):
         if self._comments is None:
@@ -45,8 +60,53 @@ class Bug(object):
         self._comments = value
 
     @classmethod
-    def search(cls, bz, args):
-        pass
+    def search(cls, bz, **kwargs):
+        """Return bugs matching the search criteria.
+
+        Values for fields are specified as keyword args.  For fields
+        with sets of valid values, appending ``"not_"`` to the field
+        negates the criterion.  If both forms are provided ("in" and
+        "not in"), the "in" criterion take precedence.
+
+        Return an Iterable of bugs (caller must not assume that the
+        value returned is a Sequence).
+        """
+        fields = frozenset([
+            'alias', 'assigned_to', 'component', 'creation_time', 'creator',
+            'id', 'last_change_time', 'op_sys', 'platform', 'priority',
+            'product', 'resolution', 'severity', 'status', 'summary',
+            'target_milestone', 'qa_contact', 'url', 'version', 'whiteboard',
+            'limit', 'offset',
+        ])
+
+        # search kwargs for "not in" args and converts to an "in",
+        # unless an "in" already exists
+        for _not_in in (k for k in kwargs if k.startswith('not_')):
+            _in = _not_in[4:]
+            if _in not in fields:
+                raise TypeError('Invalid keyword argument: {}.'.format(_in))
+            if _in not in kwargs:
+                # set _in version (_in takes precedence if it's already set)
+                if _in == 'product':
+                    all_values = set(x['name'] for x in bz.get_products())
+                else:
+                    all_values = set(
+                        value['name'] for value in
+                        [
+                            field['values'] for field in bz.get_fields()
+                            if field['name'] == _in
+                        ][0]
+                    )
+                kwargs[_in] = list(all_values - frozenset(kwargs[_not_in]))
+            del kwargs[_not_in]  # delete the _not_in
+
+        unknowns = kwargs.viewkeys() - fields
+        if unknowns:
+            # unknown arguments
+            raise TypeError(
+                'Invalid keyword arguments: {}.'.format(', '.join(unknowns)))
+        _cls = functools.partial(cls, bz)  # curry constructor with bz
+        return map(_cls, bz.rpc('Bug', 'search', **kwargs)['bugs'])
 
     def __init__(self, bz, bugno_or_data=None):
         """Create a bug object.
@@ -63,10 +123,17 @@ class Bug(object):
         self.bugno = None
         self.data = None
         self.comments = None
+        self.history = None
         try:
             self.bugno = int(bugno_or_data)
         except TypeError:
             self.data = bugno_or_data or {}
+            if 'id' in self.data:
+                self.bugno = int(self.data['id'])
+
+    @property
+    def id(self):
+        return self.bugno
 
     def rpc(self, *args, **kwargs):
         """Does an RPC on the Bugzilla server.
@@ -101,6 +168,7 @@ class Bug(object):
     def add_comment(self, comment):
         self.rpc('add_comment', id=self.bugno, comment=comment)
         self.comments = None  # comments are stale
+        self.history = None  # history is stale
 
     def is_open(self):
         """Return True if the bug is open, otherwise False."""
@@ -113,6 +181,7 @@ class Bug(object):
             kwargs['comment'] = {'body': comment}
         self.rpc('update', ids=[self.bugno], **kwargs)
         self.data = None  # data is stale
+        self.history = None  # history is stale
         if comment:
             self.comments = None  # comments are stale
 
@@ -131,6 +200,7 @@ class Bug(object):
             kwargs['comment'] = {'body': comment}
         self.rpc('update', ids=[self.bugno], **kwargs)
         self.data = None  # data is stale
+        self.history = None  # history is stale
         if comment:
             self.comments = None  # comments are stale
 
@@ -164,6 +234,7 @@ class Bug(object):
                 pass  # ignore errors (incorrect config)
         self.rpc('update', ids=[self.bugno], **kwargs)
         self.data = None  # data is stale
+        self.history = None  # history is stale
         if comment:
                 self.comments = None  # comments are stale
 
@@ -196,6 +267,7 @@ class Bug(object):
 
         result = self.rpc('update', ids=[self.bugno], **kwargs)
         self.data = None  # data is stale
+        self.history = None  # history is stale
         if 'comment' in kwargs:
             self.comments = None  # comments are stale
         return result
@@ -219,6 +291,7 @@ class Bug(object):
             kwargs['comment'] = {'body': comment}
         self.rpc('update', ids=[self.bugno], **kwargs)
         self.data = None  # data is stale
+        self.history = None  # history is stale
         if comment:
             self.comments = None  # comments are stale
 
@@ -240,6 +313,7 @@ class Bug(object):
             kwargs['comment'] = {'body': comment}
         self.rpc('update', ids=[self.bugno], **kwargs)
         self.data = None  # data is stale
+        self.history = None  # history is stale
         if comment:
             self.comments = None  # comments are stale
 
@@ -260,5 +334,20 @@ class Bug(object):
             kwargs['comment'] = {'body': comment}
         self.rpc('update', ids=[self.bugno], **kwargs)
         self.data = None  # data is stale
+        self.history = None  # history is stale
         if comment:
             self.comments = None  # comments are stale
+
+    def actual_time(self):
+        """Calculate the actual hours worked on a bug.
+
+        Hopefully this will one day be available via rpc('get', ...), but
+        for the time being, we have to use the history to calculate it.
+        """
+        changesets = (changeset['changes'] for changeset in self.history)
+        hours = (
+            float(change['added'])
+            for change in itertools.chain.from_iterable(changesets)
+            if change['field_name'] == 'work_time'
+        )
+        return sum(hours)
