@@ -1,5 +1,5 @@
 # This file is part of bugzillatools
-# Copyright (C) 2011 Benon Technologies Pty Ltd, Fraser Tweedale
+# Copyright (C) 2011, 2012 Benon Technologies Pty Ltd, Fraser Tweedale
 #
 # bugzillatools is free software: you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -18,6 +18,7 @@ from __future__ import unicode_literals
 
 import argparse
 import datetime
+import functools
 import itertools
 import re
 import textwrap
@@ -27,6 +28,8 @@ from . import bugzilla
 from . import config
 from . import editor
 
+curry = functools.partial
+
 conf = config.Config.get_config('~/.bugzillarc')
 
 
@@ -119,13 +122,9 @@ def with_limit(things='items', default=None):
 
 def with_server(cls):
     def add_server_args(parser):
-        default = None
-        if conf.has_option('core', 'server'):
-            default = conf.get('core', 'server')
         group = parser.add_argument_group('server arguments')
-        group.add_argument('--server', default=default,
-            help='Handle of Bugzilla instance to use')
-        group.add_argument('--url', help='Base URL of Bugzilla instance')
+        group.add_argument('--server', help='name of Bugzilla server to use')
+        group.add_argument('--url', help='base URL of Bugzilla server')
         group.add_argument('--user', help='Bugzilla username')
         group.add_argument('--password', help='Bugzilla password')
     cls.args = cls.args + [add_server_args]
@@ -241,22 +240,7 @@ class Help(Command):
 class BugzillaCommand(Command):
     def __init__(self, *args, **kwargs):
         super(BugzillaCommand, self).__init__(*args, **kwargs)
-        # construct the Bugzilla object
-        args = self._args
-        server = {}
-        if args.server:
-            try:
-                server = dict(conf.items('server.' + args.server))
-            except config.NoSectionError:
-                raise UserWarning(
-                    "No configuration for server '{}'.".format(args.server)
-                )
-        server.update({
-            k: getattr(args, k)
-            for k in ('url', 'user', 'password')
-            if getattr(args, k)
-        })
-        self.bz = bugzilla.Bugzilla(**server)
+        self.bz = bugzilla.Bugzilla.from_config(conf, **self._args.__dict__)
 
 
 @with_bugs
@@ -484,6 +468,36 @@ class Fields(BugzillaCommand):
                     print '  ', ','.join(map(lambda x: x['name'], values))
 
 
+def _format_history(history):
+    """Return a generator of formatted lines of history."""
+    widths = [max(len(str(h[i])) for h in history) for i in range(5)]
+    template = '{:{}}'
+    return (' | '.join(
+        (itertools.starmap(template.format, zip(map(str, h), widths)))
+    ) for h in history)
+
+
+@with_bugs
+class History(BugzillaCommand):
+    """Show the history of the given bugs."""
+    def __call__(self):
+        fields = ('WHO', 'WHEN', 'WHAT', 'REMOVED', 'ADDED')
+        for bug in map(self.bz.bug, self._args.bugs):
+            history = []
+            for h in bug.history:
+                _history = [
+                    ['', '', c['field_name'], c['removed'], c['added']]
+                    for c in h['changes']
+                ]
+                _history[0][0] = h['who']
+                _history[0][1] = h['when']
+                history.extend(_history)
+            print 'History of Bug {}:'.format(bug.bugno)
+            for line in _format_history(history):
+                print '  ' + line
+            print
+
+
 @with_bugs
 class Info(BugzillaCommand):
     """Show detailed information about the given bugs."""
@@ -504,7 +518,7 @@ class List(BugzillaCommand):
     """Show a one-line summary of the given bugs."""
     def __call__(self):
         args = self._args
-        lens = map(lambda x: len(str(x)), args.bugs)
+        lens = [len(str(x)) for x in args.bugs]
         width = max(lens) - min(lens) + 2
         for bug in map(self.bz.bug, args.bugs):
             print 'Bug {:{}} {}'.format(
@@ -523,10 +537,13 @@ class New(BugzillaCommand):
         mandatory_fields = filter(lambda x: x['is_mandatory'], fields)
 
         # first choose the product
-        b.data['product'] = self._ui.choose(
-            'Choose a product',
-            map(lambda x: x['name'], self.bz.get_products())
-        )
+        products = [x['name'] for x in self.bz.get_products()]
+        default = None
+        if 'default_product' in self.bz.config \
+                and self.bz.config['default_product'] in products:
+            default = products.index(self.bz.config['default_product'])
+        b.data['product'] = \
+            self._ui.choose('Choose a product', products, default=default)
 
         # fill out other mandatory fields
         for field in mandatory_fields:
@@ -564,10 +581,9 @@ class New(BugzillaCommand):
         ]
 
         # let user choose which of these other fields to define
-        optional_fields = [ \
-            x['display_name'] for x in fields \
-            if x['name'] in create_fields \
-                and x['name'] not in b.data.viewkeys()
+        optional_fields = [
+            x['display_name'] for x in fields
+            if x['name'] in create_fields and x['name'] not in b.data
         ]
 
         # prompt user for the fields they wish to define
@@ -593,9 +609,14 @@ class New(BugzillaCommand):
                 )
             else:
                 # TODO take field types into account
-                b.data[field['name']] = self._ui.text(
-                    'Enter the {}'.format(field['display_name'])
-                )
+                if field['name'] == 'assigned_to':
+                    _input = curry(self._ui.user, bugzilla=self.bz)
+                if field['name'] == 'cc':
+                    _input = curry(self._ui.user_list, bugzilla=self.bz)
+                else:
+                    _input = self._ui.text
+                b.data[field['name']] = \
+                    _input('Enter the {}'.format(field['display_name']))
 
         # create the bug
         id = b.create()
@@ -708,9 +729,55 @@ class Status(BugzillaCommand):
         )
 
 
-#class Search(BugzillaCommand):
-#    """Search for bugs with supplied attributes."""
-#    pass
+def _make_set_argument(arg):
+    template = 'Only match bugs {{}}of the given {}({})'.format(
+        arg, 's' if arg[-1] != 's' else 'es')
+    return [
+        lambda x: x.add_argument('--' + arg, nargs='+',
+            metavar=arg.upper(),
+            help=template.format('')),
+        lambda x: x.add_argument('--not-' + arg, nargs='+',
+            metavar=arg.upper(),
+            help=template.format('NOT ')),
+    ]
+
+
+class Search(BugzillaCommand):
+    """Search for bugs matching given criteria.
+
+    If both '--foo' and '--not-foo' are given for any argument 'foo',
+    the former takes precendence.
+    """
+    args = BugzillaCommand.args + [
+        lambda x: x.add_argument('--summary', nargs='+',
+            help='Match summary against any of the given substrings.'),
+    ]
+    simple_arguments = ['summary']
+    set_arguments = 'product', 'component', 'status', 'resolution'
+    for x in set_arguments:
+        args.extend(_make_set_argument(x))
+
+    def __call__(self):
+        kwargs = {
+            arg: getattr(self._args, arg)
+            for arg in itertools.chain(
+                self.simple_arguments,
+                self.set_arguments,
+                ('not_' + x for x in self.set_arguments)
+            )
+            if getattr(self._args, arg)
+        }
+
+        bugs = list(bug.Bug.search(self.bz, **kwargs))
+        lens = [len(str(b.bugno)) for b in bugs]
+        width = max(lens) - min(lens) + 2
+
+        for _bug in bugs:
+            print 'Bug {:{}} {}'.format(
+                str(_bug.bugno) + ':', width, _bug.data['summary']
+            )
+        n = len(bugs)
+        print '=> {} bug{} matched criteria'.format(n, 's' if n else '')
 
 
 @with_bugs
@@ -741,9 +808,9 @@ class Time(BugzillaCommand):
         else:
             # display
             #
-            # as of Bugzilla 4.0.1, "total hours worked" cannot be known.
-            # Hours are accumulated in the 'work_time' attribute of comments,
-            # which is not present in bug.comments RPC output.
+            # As of Bugzilla 4.0.1, "actual_time" (total hours worked) is
+            # not returned in bug.get.  It can, however, be calculated from
+            # the bug history.
             bugs = (self.bz.bug(bug) for bug in args.bugs)
             for bug in bugs:
                 # if user is not in the "time-tracking" group, the fields will
@@ -756,7 +823,7 @@ class Time(BugzillaCommand):
                 print '  Estimated time: {}'.format(bug.data['estimated_time'])
                 print '  Remaining time: {}'.format(bug.data['remaining_time'])
                 print '  Deadline:       {}'.format(bug.data['deadline'])
-                print '  Time worked:    {}'.format('<not available>')
+                print '  Time worked:    {}'.format(bug.actual_time())
 
 
 # the list got too long; metaprogram it ^_^
